"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const compound_1 = require("./compound");
const symbols_1 = require("./symbols");
const typeUtils_1 = require("./typeUtils");
const utils_1 = require("./utils");
const validate_1 = require("./validate");
const vality_1 = require("./vality");
vality_1.vality.array = (0, compound_1.compound)("array", (e) => (value, options, context, path) => {
    const fn = (0, utils_1.enyToGuardFn)(e);
    const { strict, bail } = (0, validate_1.mergeOptions)(options, context, ["strict", "bail"]);
    if (!Array.isArray(value)) {
        if (!strict) {
            const res = fn(value, context, [...path, 0], [value]);
            if (res.valid) {
                return { valid: true, data: [res.data], errors: [] };
            }
        }
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.array.base", path, options, value }],
        };
    }
    const data = [];
    const errors = [];
    for (let k = 0; k < value.length; k++) {
        const res = fn(value[k], context, [...path, k], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (bail)
                break;
        }
        else {
            data.push(res.data);
        }
    }
    if (errors.length === 0)
        return { valid: true, data, errors: [] };
    return { valid: false, data: undefined, errors };
}, {
    minLength: (val, o) => val.length >= o,
    maxLength: (val, o) => val.length <= o,
});
vality_1.vality.tuple = (0, compound_1.compound)("tuple", (...es) => (value, options, context, path) => {
    if (!Array.isArray(value))
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.tuple.base", path, options, value }],
        };
    const { bail } = (0, validate_1.mergeOptions)(options, context, ["bail"]);
    const data = [];
    const errors = [];
    for (let i = 0; i < es.length; i++) {
        const res = (0, utils_1.enyToGuardFn)(es[i])(value[i], context, [...path, i], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (bail)
                break;
        }
        else {
            data[i] = res.data;
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    return { valid: true, data, errors: [] };
});
vality_1.vality.object = (0, compound_1.compound)("object", (o) => (value, options, context, path, flags) => {
    if (typeof value !== "object" || value === null || Array.isArray(value))
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.object.base", path, options, value }],
        };
    const { strict, bail, allowExtraProperties } = (0, validate_1.mergeOptions)(options, context, ["strict", "bail", "allowExtraProperties"]);
    const data = {};
    const errors = [];
    for (const objectKey in o) {
        let valueKey = objectKey;
        const objectKeyEny = o[objectKey];
        if ((0, utils_1.getFlags)(objectKeyEny).has("virtual")) {
            continue;
        }
        const objectKeyGuard = (0, utils_1.enyToGuard)(objectKeyEny);
        if ((0, typeUtils_1.getName)(objectKeyGuard) === "array") {
            if (value[objectKey] === undefined &&
                value[`${objectKey}[]`] !== undefined) {
                valueKey = `${objectKey}[]`;
            }
        }
        if ((0, utils_1.getFlags)(objectKeyEny).has("from")) {
            valueKey = (0, utils_1.getFlags)(objectKeyEny).get("from");
            if (value[valueKey] === undefined) {
                if (!strict) {
                    valueKey = objectKey;
                }
            }
        }
        const res = objectKeyGuard[symbols_1._guard](value[valueKey], context, [...path, valueKey], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (bail)
                break;
        }
        else {
            data[objectKey] = res.data;
        }
    }
    if (bail && errors.length)
        return { valid: false, data: undefined, errors };
    if (!allowExtraProperties) {
        for (const valueKey in value) {
            const optionsValueEny = o[valueKey];
            if (optionsValueEny === undefined) {
                if (valueKey.endsWith("[]")) {
                    const valueKeySliced = valueKey.slice(0, -2);
                    if (value[valueKeySliced] === undefined &&
                        o[valueKeySliced] !== undefined) {
                        if ((0, typeUtils_1.getName)(o[valueKeySliced]) === "array")
                            continue;
                    }
                }
            }
            if (optionsValueEny === undefined ||
                (0, utils_1.getFlags)(optionsValueEny).has("virtual")) {
                errors.push({
                    message: "vality.object.extraProperty",
                    path: [...path, valueKey],
                    options,
                    value: value[valueKey],
                });
                if (bail)
                    break;
            }
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    return { valid: true, data, errors: [] };
});
vality_1.vality.dict = (0, compound_1.compound)("dict", (k, v) => (value, options, context, path) => {
    if (typeof value !== "object" || value === null) {
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.dict.base", path, options, value }],
        };
    }
    if ((0, utils_1.getFlags)(v).has("virtual")) {
        return { valid: true, data: undefined, errors: [] };
    }
    const { bail, allowExtraProperties } = (0, validate_1.mergeOptions)(options, context, ["bail", "allowExtraProperties"]);
    const type = (0, typeUtils_1.getName)(k);
    let literalKeys;
    let typeKeys;
    switch (type) {
        case "literal":
            literalKeys = [k];
            typeKeys = [];
            break;
        case "enum":
            const simpleKeyGuard = (0, typeUtils_1.simplifyEnum)(k[symbols_1._type]);
            [literalKeys, typeKeys] = simpleKeyGuard.reduce((acc, key) => {
                const guard = (0, utils_1.enyToGuard)(key);
                const type = (0, typeUtils_1.getName)(guard);
                if (type === "literal") {
                    acc[0].push(guard);
                }
                else {
                    acc[1].push(guard);
                }
                return acc;
            }, [[], []]);
            break;
        default:
            literalKeys = [];
            typeKeys = [k];
            break;
    }
    const valueGuard = (0, utils_1.enyToGuard)(v);
    const data = {};
    const errors = [];
    const seenKeys = new Set();
    for (const literal of literalKeys) {
        const literalGuard = (0, utils_1.enyToGuard)(literal);
        const literalValue = literalGuard[symbols_1._type][0][symbols_1._type];
        const res = valueGuard[symbols_1._guard](value[literalValue], context, [...path, literalValue], value);
        if (!res.valid) {
            errors.push(...res.errors);
            if (bail)
                break;
        }
        else {
            seenKeys.add(literalValue);
            data[literalValue] = res.data;
        }
    }
    if (bail && errors.length)
        return { valid: false, data: undefined, errors };
    for (const valueKey in value) {
        if (seenKeys.has(valueKey))
            continue;
        let remappedKey;
        for (const typeKey of typeKeys) {
            const res = typeKey[symbols_1._guard](valueKey, context, [...path, valueKey], value);
            if (res.valid) {
                remappedKey = res.data;
                break;
            }
        }
        if (remappedKey === undefined) {
            if (!allowExtraProperties) {
                errors.push({
                    message: "vality.dict.extraProperty",
                    path: [...path, valueKey],
                    options,
                    value: value[valueKey],
                });
                if (bail)
                    break;
            }
        }
        else {
            const res = valueGuard[symbols_1._guard](value[valueKey], context, [...path, valueKey], value);
            if (!res.valid) {
                errors.push(...res.errors);
                if (bail)
                    break;
            }
            else {
                data[remappedKey] = res.data;
            }
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    return { valid: true, data, errors: [] };
});
vality_1.vality.enum = (0, compound_1.compound)("enum", (...es) => (value, options, context, path, parent) => {
    for (const e of es) {
        const res = (0, utils_1.enyToGuardFn)(e)(value, context, path, parent);
        if (res.valid)
            return res;
    }
    return {
        valid: false,
        data: undefined,
        errors: [{ message: "vality.enum.base", path, options, value }],
    };
});
vality_1.vality.and = (0, compound_1.compound)("and", (...es) => (value, options, context, path, parent) => {
    if (typeof value !== "object" || value === null)
        return {
            valid: false,
            data: undefined,
            errors: [{ message: "vality.and.base", path, options, value }],
        };
    const { bail } = (0, validate_1.mergeOptions)(options, context, ["bail"]);
    const data = {};
    const errors = [];
    for (const e of (0, typeUtils_1.intersect)(es)) {
        const res = (0, utils_1.enyToGuardFn)(e)(value, context, path, parent);
        if (!res.valid) {
            errors.push(...res.errors);
            if (bail)
                break;
        }
        else {
            Object.assign(data, res.data);
        }
    }
    if (errors.length)
        return { valid: false, data: undefined, errors };
    return { valid: true, data, errors: [] };
});
//# sourceMappingURL=compounds.js.map