"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const scalar_1 = require("./scalar");
const symbols_1 = require("./symbols");
const validate_1 = require("./validate");
const vality_1 = require("./vality");
vality_1.vality.string = (0, scalar_1.scalar)("string", (val, options, context) => {
    if (typeof val === "string")
        return val;
    const { strict } = (0, validate_1.mergeOptions)(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "number")
        return undefined;
    return val.toString();
}, {
    minLength: (val, o) => val.length >= o,
    maxLength: (val, o) => val.length <= o,
    match: (val, o) => o.test(val),
});
vality_1.vality.number = (0, scalar_1.scalar)("number", (val, options, context) => {
    if (typeof val === "number")
        return val;
    const { strict } = (0, validate_1.mergeOptions)(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "string")
        return undefined;
    const nr = Number.parseFloat(val);
    if (Number.isNaN(nr))
        return undefined;
    return nr;
}, {
    min: (val, o) => val >= o,
    max: (val, o) => val <= o,
    onlyInteger: (val, o) => !o || val % 1 === 0,
    allowUnsafe: (val, o) => o ||
        (val > Number.MIN_SAFE_INTEGER && val < -Number.EPSILON) ||
        (val < Number.MAX_SAFE_INTEGER && val > Number.EPSILON) ||
        val === 0,
    allowNaN: (val, o) => o || !Number.isNaN(val),
}, {
    onlyInteger: false,
    allowUnsafe: false,
    allowNaN: false,
});
const y = new Set().add("1").add(1).add("true");
const n = new Set().add("0").add(0).add("false");
vality_1.vality.boolean = (0, scalar_1.scalar)("boolean", (val, options, context) => {
    if (typeof val === "boolean")
        return val;
    const { strict } = (0, validate_1.mergeOptions)(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "string" && typeof val !== "number")
        return undefined;
    return y.has(val) ? true : n.has(val) ? false : undefined;
});
vality_1.vality.date = (0, scalar_1.scalar)("date", (val, options, context) => {
    if (val instanceof Date)
        return val;
    const { strict } = (0, validate_1.mergeOptions)(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "string" && typeof val !== "number")
        return undefined;
    const date = new Date(val);
    if (Number.isNaN(date.getTime()))
        return undefined;
    return date;
}, {
    min: (val, o) => val >= o,
    max: (val, o) => val <= o,
    past: (val, o) => !o || val < new Date(),
    future: (val, o) => !o || val >= new Date(),
}, {
    past: false,
    future: false,
});
vality_1.vality.any = (0, scalar_1.scalar)("any", (val) => val);
vality_1.vality.never = (0, scalar_1.scalar)("never", () => undefined);
vality_1.vality.literal = (lit) => {
    const scalarFn = (val, options, context) => {
        if (options.default === true) {
            if (val === undefined)
                return lit;
        }
        else {
            delete options.default;
        }
        if (val === lit)
            return lit;
        const { strict } = (0, validate_1.mergeOptions)(options, context, ["strict"]);
        if (strict)
            return undefined;
        if (typeof lit === "string" && typeof val === "number")
            return val.toString() === lit ? lit : undefined;
        if (typeof lit === "number" && typeof val === "string")
            return parseFloat(val) === lit ? lit : undefined;
        if (typeof lit === "boolean" &&
            (typeof val === "string" || typeof val === "number")) {
            if (lit === true && y.has(val))
                return lit;
            if (lit === false && n.has(val))
                return lit;
        }
        return undefined;
    };
    scalarFn[symbols_1._type] = lit;
    return (0, scalar_1.scalar)("literal", scalarFn);
};
//# sourceMappingURL=scalars.js.map