"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getName = exports.simplifyEnum = exports.intersect = void 0;
const symbols_1 = require("./symbols");
const types_1 = require("./types");
const utils_1 = require("./utils");
const vality_1 = require("./vality");
function intersect(objs) {
    if (!objs.length)
        return [vality_1.vality.never];
    let values = [].concat(...objs.map((member) => {
        if (getName(member) === "and")
            return intersect(member[symbols_1._type]);
        return (0, utils_1.enyToGuard)(member);
    }));
    values = values.filter(v => !isAny(v));
    if (values.length === 0)
        return [vality_1.vality.any];
    if (values.some(isNever))
        return [vality_1.vality.never];
    if (values.length === 1)
        return values;
    return values;
}
exports.intersect = intersect;
function simplifyEnum(objs) {
    if (!objs.length)
        return [vality_1.vality.never];
    let values = [].concat(...objs
        .map((member) => {
        if (getName(member) === "enum")
            return simplifyEnum(member[symbols_1._type]);
        return (0, utils_1.enyToGuard)(member);
    }));
    values = values.filter((v) => !isNever(v));
    if (values.length === 0)
        return [vality_1.vality.never];
    if (values.some(isAny))
        return [vality_1.vality.any];
    if (values.length === 1)
        return values;
    let hasStringType = false;
    let hasNumberType = false;
    for (const value of values) {
        if (getName(value) === "string")
            hasStringType = true;
        if (getName(value) === "number")
            hasNumberType = true;
        if (hasStringType && hasNumberType)
            return [vality_1.vality.string, vality_1.vality.number];
    }
    values = values.filter((v) => {
        if (hasStringType &&
            getName(v) === "literal" &&
            typeof v[symbols_1._type][0][symbols_1._type] === "string")
            return false;
        if (hasNumberType &&
            getName(v) === "literal" &&
            typeof v[symbols_1._type][0][symbols_1._type] === "number")
            return false;
        return true;
    });
    return values;
}
exports.simplifyEnum = simplifyEnum;
function isNever(g) {
    return getName(g) === "never";
}
function isAny(g) {
    return getName(g) === "any";
}
function getName(e) {
    const g = (0, utils_1.enyToGuard)(e);
    let type = g[symbols_1._name];
    const seen = new Set();
    while (types_1.types.has(type)) {
        type = types_1.types.get(type);
        if (seen.has(type))
            throw new Error("Circular type extension");
        seen.add(type);
    }
    return type;
}
exports.getName = getName;
//# sourceMappingURL=typeUtils.js.map