import { _name, _type } from "./symbols";
import { types } from "./types";
import { enyToGuard } from "./utils";
import { vality } from "./vality";
export function intersect(objs) {
    if (!objs.length)
        return [vality.never];
    let values = [].concat(...objs.map((member) => {
        if (getName(member) === "and")
            return intersect(member[_type]);
        return enyToGuard(member);
    }));
    values = values.filter(v => !isAny(v));
    if (values.length === 0)
        return [vality.any];
    if (values.some(isNever))
        return [vality.never];
    if (values.length === 1)
        return values;
    return values;
}
export function simplifyEnum(objs) {
    if (!objs.length)
        return [vality.never];
    let values = [].concat(...objs
        .map((member) => {
        if (getName(member) === "enum")
            return simplifyEnum(member[_type]);
        return enyToGuard(member);
    }));
    values = values.filter((v) => !isNever(v));
    if (values.length === 0)
        return [vality.never];
    if (values.some(isAny))
        return [vality.any];
    if (values.length === 1)
        return values;
    let hasStringType = false;
    let hasNumberType = false;
    for (const value of values) {
        if (getName(value) === "string")
            hasStringType = true;
        if (getName(value) === "number")
            hasNumberType = true;
        if (hasStringType && hasNumberType)
            return [vality.string, vality.number];
    }
    values = values.filter((v) => {
        if (hasStringType &&
            getName(v) === "literal" &&
            typeof v[_type][0][_type] === "string")
            return false;
        if (hasNumberType &&
            getName(v) === "literal" &&
            typeof v[_type][0][_type] === "number")
            return false;
        return true;
    });
    return values;
}
function isNever(g) {
    return getName(g) === "never";
}
function isAny(g) {
    return getName(g) === "any";
}
export function getName(e) {
    const g = enyToGuard(e);
    let type = g[_name];
    const seen = new Set();
    while (types.has(type)) {
        type = types.get(type);
        if (seen.has(type))
            throw new Error("Circular type extension");
        seen.add(type);
    }
    return type;
}
//# sourceMappingURL=typeUtils.js.map