import { _flags, _guard, _name, _type } from "./symbols";
export function makeValit(...[name, fn, handleOptions, defaultOptions]) {
    return (...args) => {
        const getGuardFnFromOptions = (options) => (value, context, path, parent) => {
            const { transform, preprocess, default: defaultValue, validate, bail, ...optionsWithoutExtras } = options;
            const optionsWithDefault = {
                ...defaultOptions,
                ...optionsWithoutExtras,
            };
            if (typeof preprocess === "function") {
                value = preprocess(value, options, context, path, parent);
            }
            const data = fn(...args)(value, options, context, path, parent);
            if (!data.valid) {
                if (value === undefined && defaultValue !== undefined) {
                    return { valid: true, data: defaultValue, errors: [] };
                }
                return data;
            }
            if (typeof validate === "function" && validate(data.data, options, context, path, parent) === false) {
                return {
                    valid: false,
                    data: undefined,
                    errors: [
                        {
                            message: `vality.${name}.custom`,
                            path,
                            options,
                            value,
                        },
                    ],
                };
            }
            const origData = data.data;
            if (typeof transform === "function") {
                data.data = transform(data.data);
            }
            if (handleOptions === undefined)
                return data;
            const keysWithError = [];
            for (const key in optionsWithDefault) {
                if (handleOptions[key]?.(origData, optionsWithDefault[key], options, context) === false) {
                    keysWithError.push(key);
                    if (bail)
                        break;
                }
            }
            if (keysWithError.length === 0)
                return data;
            return {
                valid: false,
                data: undefined,
                errors: keysWithError.map((k) => ({
                    message: k in options
                        ? `vality.${name}.options.${k}`
                        : `vality.${name}.base`,
                    options,
                    path,
                    value,
                })),
            };
        };
        const flagsMap = new Map();
        const validate = (options => ({
            [_name]: name,
            [_guard]: (val, context, path, parent) => {
                if (typeof options === "function")
                    options = options(parent, context);
                return getGuardFnFromOptions(options)(val, context, path, parent);
            },
            [_type]: args,
            [_flags]: flagsMap,
        }));
        validate[_name] = name;
        validate[_guard] = getGuardFnFromOptions({});
        validate[_type] = args;
        validate[_flags] = flagsMap;
        return validate;
    };
}
//# sourceMappingURL=valit.js.map